"""ACA-Py Pickup Protocol Plugin."""

from aries_cloudagent.config.injection_context import InjectionContext
from aries_cloudagent.core.protocol_registry import ProtocolRegistry
from .protocol.live_mode import LiveDeliveryChange
from .protocol.delivery import DeliveryRequest, Delivery, MessagesReceived
from .protocol.status import StatusRequest, Status


async def setup(context: InjectionContext):
    """Setup plugin."""
    protocol_registry = context.inject(ProtocolRegistry)
    assert protocol_registry
    protocol_registry.register_message_types(
        {
            Status.message_type: Status,
            StatusRequest.message_type: StatusRequest,
            Delivery.message_type: Delivery,
            DeliveryRequest.message_type: DeliveryRequest,
            MessagesReceived.message_type: MessagesReceived,
            LiveDeliveryChange.message_type: LiveDeliveryChange,
        }
    )



from aries_cloudagent.messaging.request_context import RequestContext
from aries_cloudagent.messaging.responder import BaseResponder

from ..acapy import AgentMessage

PROTOCOL = "https://didcomm.org/messagepickup/2.0"


# This is the start of a message updating the Live Delivery status
# Will require a deeper analysis of ACA-Py to fully implement
class LiveDeliveryChange(AgentMessage):
    """Live Delivery Change message."""

    message_type = f"{PROTOCOL}/live-delivery-change"

    live_delivery: bool = False

    async def handle(self, context: RequestContext, responder: BaseResponder):
        """Handle LiveDeliveryChange message"""
        return await super().handle(context, responder)



"""Status Request and Status messages for the Pickup Protocol."""

import logging
from typing import Optional

from aries_cloudagent.messaging.request_context import RequestContext
from aries_cloudagent.messaging.responder import BaseResponder
from aries_cloudagent.transport.inbound.manager import InboundTransportManager

from ..acapy import AgentMessage
from ..acapy.error import HandlerException
from ..valid import ISODateTime

LOGGER = logging.getLogger(__name__)
PROTOCOL = "https://didcomm.org/messagepickup/2.0"


class StatusRequest(AgentMessage):
    """StatusRequest message."""

    message_type = f"{PROTOCOL}/status-request"

    recipient_key: Optional[str] = None

    async def handle(self, context: RequestContext, responder: BaseResponder):
        """Handle status request message."""
        if not self.transport or self.transport.return_route != "all":
            raise HandlerException(
                "StatusRequest must have transport decorator with return "
                "route set to all"
            )
        recipient_key = self.recipient_key
        manager = context.inject(InboundTransportManager)
        assert manager
        count = manager.undelivered_queue.message_count_for_key(
            recipient_key or context.message_receipt.sender_verkey
        )
        response = Status(message_count=count, recipient_key=recipient_key)
        response.assign_thread_from(self)
        await responder.send_reply(response)


class Status(AgentMessage):
    """Status message."""

    message_type = f"{PROTOCOL}/status"

    message_count: int
    recipient_key: Optional[str] = None
    duration_waited: Optional[int] = None
    newest_time: Optional[ISODateTime] = None
    oldest_time: Optional[ISODateTime] = None
    total_size: Optional[int] = None
    live_mode: Optional[bool] = None



"""Simple Agent Message class."""

from abc import ABC
from datetime import datetime
import json
import logging
from typing import Any, ClassVar, Dict, Mapping, Optional, Union
from typing_extensions import Annotated, Literal
from uuid import uuid4
import uuid

from aries_cloudagent.messaging.base_handler import BaseHandler
from aries_cloudagent.messaging.base_message import BaseMessage
from aries_cloudagent.messaging.request_context import RequestContext
from aries_cloudagent.messaging.responder import BaseResponder
from aries_cloudagent.wallet.util import bytes_to_b64
from pydantic import BaseModel, Field, parse_obj_as
from pydantic.class_validators import validator, root_validator
from pydantic.types import StrictInt

from acapy_plugin_pickup.valid import ISODateTime


LOGGER = logging.getLogger(__name__)


class AttachData(BaseModel):
    class Config:
        allow_population_by_field_name = True

    base64: Annotated[Optional[str], Field(description="Base64-encoded data")] = None
    json_: Annotated[
        Optional[dict], Field(description="JSON-serialized data", alias="json")
    ] = None
    links: Annotated[
        Optional[str], Field(description="List of hypertext links to data")
    ] = None

    @root_validator(pre=True)
    @classmethod
    def _mutual_exclusion_validate(cls, values):
        if len(set(values.keys()) & {"base64", "json", "links", "json_"}) != 1:
            raise ValueError("AttachData: choose exactly one of base64, json, or links")
        return values


class Attach(BaseModel):
    ident: Annotated[str, Field(alias="@id")]
    mime_type: Annotated[str, Field(alias="mime-type")]
    filename: Optional[str] = None
    byte_count: Optional[StrictInt] = None
    lastmod_time: Optional[datetime]
    description: Optional[str]
    data: AttachData

    class Config:
        allow_population_by_field_name = True

    @classmethod
    def data_base64(
        cls,
        value: Union[Mapping, bytes, str],
        *,
        ident: str = None,
        description: str = None,
        filename: str = None,
        lastmod_time: str = None,
        byte_count: int = None,
    ):
        """
        Create `AttachDecorator` instance on base64-encoded data from input mapping.

        Given mapping, JSON dump, base64-encode, and embed
        it as data; mark `application/json` MIME type.

        Args:
            mapping: (dict) data structure; e.g., indy production
            ident: optional attachment identifier (default random UUID4)
            description: optional attachment description
            filename: optional attachment filename
            lastmod_time: optional attachment last modification time
            byte_count: optional attachment byte count

        """
        if isinstance(value, Mapping):
            value = json.dumps(value).encode()
        if isinstance(value, str):
            value = value.encode()

        attach_data = AttachData(base64=bytes_to_b64(value))

        return cls(
            ident=ident or str(uuid.uuid4()),
            description=description,
            filename=filename,
            mime_type="application/json",
            lastmod_time=lastmod_time,
            byte_count=byte_count,
            data=attach_data,
        )


class Thread(BaseModel):
    thid: Annotated[
        Optional[str],
        Field(
            description="Thread identifier",
            examples=[uuid4()],  # typically a UUID4 but not necessarily
        ),
    ] = None
    pthid: Annotated[
        Optional[str],
        Field(
            description="Parent thread identifier",
            examples=[uuid4()],  # typically a UUID4 but not necessarily
        ),
    ] = None
    sender_order: Annotated[
        Optional[int],
        Field(
            description="Ordinal of message among all from current sender in thread",
            examples=[11, 0, 2],
        ),
    ] = None
    received_orders: Annotated[
        Optional[
            Dict[
                Annotated[str, Field(description="Sender key")],
                Annotated[
                    int, Field(description="Highest sender_order value for sender")
                ],
            ]
        ],
        Field(
            description="Highest sender_order value that sender has seen from "
            "others on thread"
        ),
    ] = None


class Transport(BaseModel):
    """Tranpsport Decorator object."""

    return_route: Literal["none", "all", "thread"] = "none"


class AgentMessage(BaseModel, BaseMessage, ABC):
    """AgentMessage Interface Definition."""

    message_type: ClassVar[str] = ""

    id: Annotated[str, Field(alias="@id", default_factory=lambda: str(uuid4()))]
    type: Annotated[Optional[str], Field(alias="@type")] = None
    thread: Annotated[Optional[Thread], Field(alias="~thread")] = None
    transport: Annotated[Optional[Transport], Field(alias="~transport")] = None

    class Config:
        """AgentMessage Config."""

        json_encoders = {ISODateTime: lambda value: value.isoformat()}

    @validator("type", pre=True, always=True)
    @classmethod
    def _type(cls, value):
        """Set type if not present."""
        if not value:
            return cls.message_type
        if value != cls.message_type:
            raise ValueError(
                "Invalid message type for {}: {}".format(cls.__name__, value)
            )
        return value

    @property
    def _id(self):
        return self.id

    @property
    def _thread_id(self) -> Optional[str]:
        """Return this message's thread id."""
        return self.thread and self.thread.thid

    def serialize(self) -> dict:
        """Serialize an instance of message to dictionary."""
        return self.dict(exclude_none=True, by_alias=True)

    @classmethod
    def deserialize(cls, value: Mapping[str, Any]) -> "AgentMessage":
        """Deserialize an instance of message."""
        return parse_obj_as(cls, value)

    def assign_thread_from(self, msg: "AgentMessage"):
        """Assign thread info from another message."""
        if msg:
            thread = msg.thread
            thid = thread.thid if thread else msg.id
            pthid = thread and thread.pthid
            self.assign_thread_id(thid, pthid)

    def assign_thread_id(self, thid: str = None, pthid: str = None):
        """Assign thread info."""
        if thid or pthid:
            self.thread = Thread(thid=thid, pthid=pthid)
        else:
            self.thread = None

    def json(self, **kwargs):
        """Dump to json."""
        return super().json(exclude_none=True, by_alias=True, **kwargs)

    async def handle(self, context: RequestContext, responder: BaseResponder):
        """Handle a message of this type."""
        LOGGER.debug("Received message of type %s:\n%s", self.type, self.json(indent=2))

    @property
    def Handler(self):
        msg_class = self.__class__

        class Handler(BaseHandler):
            """Handler for message."""

            async def handle(self, context, responder):
                """Handle the message."""
                return await msg_class.handle(context.message, context, responder)

        return Handler

    # Fulfill Responder Message Protocol

    def to_json(self) -> str:
        """Dump to json."""
        return self.json()



"""Validation helpers."""

from datetime import datetime
from dateutil import parser
from pydantic.class_validators import validator


class ISODateTime(datetime):
    """Custom field for ISO formatted datetime."""

    @classmethod
    def __get_validators__(cls):
        """Yield validators."""
        yield cls.validate

    @validator("__root__", pre=True)
    @classmethod
    def validate(cls, value):
        """Validate the datetime value as ISO time format."""
        return parser.isoparse(value)



"""Delivery Request and wrapper message for Pickup Protocol."""

import json
import logging
from typing import List, Optional, Sequence, Set, cast

from aries_cloudagent.messaging.request_context import RequestContext
from aries_cloudagent.messaging.responder import BaseResponder
from aries_cloudagent.transport.inbound.delivery_queue import DeliveryQueue
from aries_cloudagent.transport.inbound.manager import InboundTransportManager
from aries_cloudagent.transport.inbound.session import InboundSession
from aries_cloudagent.transport.outbound.message import OutboundMessage
from aries_cloudagent.transport.wire_format import BaseWireFormat
from pydantic import Field
from typing_extensions import Annotated

from ..acapy import AgentMessage, Attach
from ..acapy.error import HandlerException
from .status import Status

LOGGER = logging.getLogger(__name__)
PROTOCOL = "https://didcomm.org/messagepickup/2.0"


class DeliveryRequest(AgentMessage):
    """DeliveryRequest message."""

    message_type = f"{PROTOCOL}/delivery-request"

    limit: int
    recipient_key: Optional[str] = None

    @staticmethod
    def determine_session(manager: InboundTransportManager, key: str):
        """Determine the session associated with the given key."""
        for session in manager.sessions.values():
            session = cast(InboundSession, session)
            if key in session.reply_verkeys:
                return session
        return None

    async def handle(self, context: RequestContext, responder: BaseResponder):
        """Handle DeliveryRequest message."""
        if not self.transport or self.transport.return_route != "all":
            raise HandlerException(
                "DeliveryRequest must have transport decorator with return "
                "route set to all"
            )

        wire_format = context.inject(BaseWireFormat)
        manager = context.inject(InboundTransportManager)
        assert manager
        queue = manager.undelivered_queue
        key = context.message_receipt.sender_verkey
        message_attachments = []

        if queue.has_message_for_key(key):
            session = self.determine_session(manager, key)
            if session is None:
                LOGGER.warning("No session available to deliver messages as requested")
                return

            returned_count = 0
            async with context.session() as profile_session:

                for msg in get_messages_for_key(queue, key):

                    recipient_key = (
                        msg.target_list[0].recipient_keys
                        or context.message_receipt.recipient_verkey
                    )
                    routing_keys = msg.target_list[0].routing_keys or []
                    sender_key = msg.target_list[0].sender_key or key

                    # This scenario is rare; a message will almost always have an
                    # encrypted payload. The only time it won't is if we're sending a
                    # message from the mediator itself, rather than forwarding a message
                    # from another agent.
                    # TODO: update ACA-Py to store all messages with an
                    # encrypted payload
                    if not msg.enc_payload:
                        msg.enc_payload = await wire_format.encode_message(
                            profile_session,
                            msg.payload,
                            recipient_key,
                            routing_keys,
                            sender_key,
                        )

                    attached_msg = Attach.data_base64(
                        ident=json.loads(msg.enc_payload)["tag"], value=msg.enc_payload
                    )
                    message_attachments.append(attached_msg)
                    returned_count += 1

                    if returned_count >= self.limit:
                        break

            response = Delivery(message_attachments=message_attachments)
            response.assign_thread_from(self)
            await responder.send_reply(response)

        else:
            response = Status(recipient_key=self.recipient_key, message_count=0)
            response.assign_thread_from(self)
            await responder.send_reply(response)


class Delivery(AgentMessage):
    """Message wrapper for delivering messages to a recipient."""

    class Config:
        allow_population_by_field_name = True

    message_type = f"{PROTOCOL}/delivery"

    recipient_key: Optional[str] = None
    message_attachments: Annotated[
        Sequence[Attach], Field(description="Attached messages", alias="~attach")
    ]


class MessagesReceived(AgentMessage):
    """MessageReceived acknowledgement message."""

    message_type = f"{PROTOCOL}/messages-received"
    message_id_list: Set[str]

    async def handle(self, context: RequestContext, responder: BaseResponder):
        """Handle MessageReceived message."""
        if not self.transport or self.transport.return_route != "all":
            raise HandlerException(
                "MessageReceived must have transport decorator with return "
                "route set to all"
            )

        manager = context.inject(InboundTransportManager)
        assert manager
        queue = manager.undelivered_queue
        key = context.message_receipt.sender_verkey

        if queue.has_message_for_key(key):
            remove_message_by_tag_list(queue, key, self.message_id_list)

        response = Status(message_count=queue.message_count_for_key(key))
        response.assign_thread_from(self)
        await responder.send_reply(response)


def remove_message_by_tag(queue: DeliveryQueue, recipient_key: str, tag: str):
    """Remove a message from a recipient's queue by tag.

    Tag corresponds to a value in the encrypted payload which is unique for
    each message.
    """
    return remove_message_by_tag_list(queue, recipient_key, {tag})


def remove_message_by_tag_list(
    queue: DeliveryQueue, recipient_key: str, tag_list: Set[str]
):

    # For debugging, logs the contents of each message as it's retrieved from the queue
    for i in queue.queue_by_key[recipient_key]:
        LOGGER.debug("%s", i.msg)

    if recipient_key not in queue.queue_by_key:
        return
    LOGGER.debug(
        "Removing messages with tags from queue: %s", queue.queue_by_key[recipient_key]
    )
    queue.queue_by_key[recipient_key][:] = [
        queued_message
        for queued_message in queue.queue_by_key[recipient_key]
        if queued_message.msg.enc_payload is None
        or json.loads(queued_message.msg.enc_payload)["tag"] not in tag_list
    ]


def get_messages_for_key(queue: DeliveryQueue, key: str) -> List[OutboundMessage]:
    """
    Return messages for a given key from the queue without removing them.

    Args:
        key: The key to use for lookup
    """
    if key in queue.queue_by_key:
        return [queued.msg for queued in queue.queue_by_key[key]]
    return []



# -*- coding: utf-8 -*-
from setuptools import setup

packages = \
['acapy_plugin_pickup',
 'acapy_plugin_pickup.acapy',
 'acapy_plugin_pickup.protocol']

package_data = \
{'': ['*']}

install_requires = \
['aries-cloudagent>=0.7.3,<0.8.0',
 'pydantic>=1.8.1,<2.0.0',
 'python-dateutil>=2.8.1,<3.0.0',
 'typing-extensions>=3.7.4,<4.0.0']

extras_require = \
{'indy': ['python3-indy>=1.11.1']}

setup_kwargs = {
    'name': 'acapy-plugin-pickup',
    'version': '0.1.0.post1',
    'description': 'Pickup Protocol for ACA-Py',
    'long_description': None,
    'author': 'Daniel Bluhm',
    'author_email': 'dbluhm@pm.me',
    'maintainer': None,
    'maintainer_email': None,
    'url': None,
    'packages': packages,
    'package_data': package_data,
    'install_requires': install_requires,
    'extras_require': extras_require,
    'python_requires': '>=3.6.9,<4.0',
}


setup(**setup_kwargs)
